"""Provides shared constants for Base Callables"""

from nawah.classes import Attr

FILE_ATTRS = {
    "user": Attr.ID(),
    "doc": Attr.ID(),
    "attr": Attr.STR(),
    "file": Attr.ANY(),
    "create_time": Attr.DATETIME(),
}



"""Provides 'check_permissions' Utility"""

import copy
import logging
from typing import TYPE_CHECKING, Optional, cast

from nawah.exceptions import NotPermittedException

if TYPE_CHECKING:
    from nawah.classes import Func, Module, QueryMod
    from nawah.types import NawahDoc, NawahEnv

logger = logging.getLogger("nawah")


def check_permissions(
    *, func: "Func", env: "NawahEnv"
) -> tuple[Optional["QueryMod"], Optional["NawahDoc"]]:
    """Matches Permission Sets against current session user privileges. Returns 'Func' object
    query_mod, doc_mod of Permission Set matched. If failed to match any, raises"""

    module = cast("Module", func.module)

    logger.debug(
        "Attempting to check permissions for '%s'.'%s' against '%s'",
        module.name,
        func.name,
        env["session"]["user"],
    )

    # Loop over Permissions Set to find a match and return its query_mod, doc_mod
    for permission in func.permissions:
        # Check for allow-all *
        if permission.privilege == "*":
            return (
                copy.deepcopy(permission.query_mod),
                copy.deepcopy(permission.doc_mod),
            )
        # Set variables to use for checking permission
        privilege_module = module.name
        privilege = permission.privilege
        # If Permission Set privilege is dot-notated value, use it to set module, privilege
        if "." in permission.privilege:
            privilege_module, privilege = permission.privilege.split(".")

        # Skip Permission Set if privilege_module not in user prvileges
        if privilege_module not in env["session"]["user"]["privileges"]:
            continue

        if privilege not in env["session"]["user"]["privileges"][privilege_module]:
            continue

        return (copy.deepcopy(permission.query_mod), copy.deepcopy(permission.doc_mod))

    # If no match, raise NotPermittedException
    raise NotPermittedException(module_name=module.name, func_name=func.name)



"""Provides 'base' Nawah Core Module"""

from nawah.classes import Attr, Func, Module, Perm

base = Module(
    name="base",
    desc="'base' module provides necessary structure to allow modules to access Base "
    "Functions Callables via 'call' Utility",
    funcs={
        "read": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "create": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "update": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "delete": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "obtain_lock": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            doc_attrs={"tags": Attr.LIST(list=[Attr.STR()]), "attempts": Attr.INT()},
        ),
        "delete_lock": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
    },
)



"""Provides collection of validation tools used by Nawah and
can be used by Nawah Apps to use with complex business-logic"""

from ._attr import validate_attr, validate_doc
from ._type import validate_type

__all__ = ["validate_attr", "validate_doc", "validate_type"]



"""Provides 'user' Nawah Core Module"""

from nawah.classes import Attr, Default, Func, Module, Perm, Var

from ._callables import (_add_group, _create, _delete_group, _read,
                         _read_privileges)
from ._exceptions import (GroupAddedException, GroupNotAddedException,
                          InvalidGroupException, InvalidUserException)

user = Module(
    name="user",
    desc="'user' module provides data type and controller for users in Nawah eco-system. The "
    "permissions of the module methods are designed to be as secure for exposed calls, and as "
    "flexible for privileged-access",
    collection="user_docs",
    attrs={
        "name": Attr.LOCALE(desc="Name of the user as 'LOCALE'"),
        "locale": Attr.LOCALES(desc="Default locale of the user"),
        "create_time": Attr.DATETIME(
            desc="Python 'datetime' ISO format of the doc creation"
        ),
        "login_time": Attr.DATETIME(
            desc="Python 'datetime' ISO format of the last login"
        ),
        "groups": Attr.LIST(
            desc="List of '_id' for every group the user is member of",
            list=[Attr.ID(desc="'_id' of Group doc the user is member of")],
        ),
        "privileges": Attr.KV_DICT(
            desc="Privileges of the user. These privileges are always available to the user"
            "regardless of whether groups user is part of have them or not",
            key=Attr.STR(),
            val=Attr.LIST(list=[Attr.STR()]),
        ),
        "status": Attr.LITERAL(
            desc="Status of the user to determine whether user has access to the app or not",
            literal=["active", "banned", "deleted", "disabled_password"],
        ),
    },
    defaults={
        "login_time": Default(value=None),
        "status": Default(value="active"),
        "groups": Default(value=[]),
        "privileges": Default(value={}),
    },
    funcs={
        "read": Func(
            permissions=[
                Perm(privilege="admin"),
                Perm(privilege="read", query_mod={"_id": Var.SESSION("user._id")}),
            ],
            call_args={
                "skip_sanitise_results": Attr.BOOL(),
                "raise_no_success": Attr.BOOL(),
            },
            callable=_read,
        ),
        "create": Func(permissions=[Perm(privilege="admin")], callable=_create),
        "update": Func(
            permissions=[
                Perm(privilege="admin", doc_mod={"groups": None}),
                Perm(
                    privilege="update",
                    query_mod={"_id": Var.SESSION("user._id")},
                    doc_mod={"groups": None, "privileges": None},
                ),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
        "delete": Func(
            permissions=[
                Perm(privilege="admin"),
                Perm(privilege="delete", query_mod={"_id": Var.SESSION("user._id")}),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
        "read_privileges": Func(
            permissions=[
                Perm(privilege="admin"),
                Perm(privilege="read", query_mod={"_id": Var.SESSION("user._id")}),
            ],
            query_attrs={"_id": Attr.ID()},
            callable=_read_privileges,
            exceptions={
                InvalidUserException: True,
            },
        ),
        "add_group": Func(
            permissions=[Perm(privilege="admin")],
            query_attrs={"_id": Attr.ID()},
            doc_attrs=[{"group": Attr.ID()}, {"group": Attr.LIST(list=[Attr.ID()])}],
            callable=_add_group,
            exceptions={
                InvalidGroupException: False,
                InvalidUserException: False,
                GroupAddedException: False,
            },
        ),
        "delete_group": Func(
            permissions=[Perm(privilege="admin")],
            query_attrs={"_id": Attr.ID(), "group": Attr.ID()},
            callable=_delete_group,
            exceptions={
                InvalidGroupException: False,
                InvalidUserException: False,
                GroupNotAddedException: False,
            },
        ),
        "retrieve_file": Func(permissions=[Perm(privilege="__sys")], get_func=True),
        "create_file": Func(permissions=[Perm(privilege="__sys")]),
        "update_file": Func(permissions=[Perm(privilege="__sys")]),
        "delete_file": Func(permissions=[Perm(privilege="__sys")]),
    },
)



"""Provides classes used in Nawah"""

from ._analytics import Analytics
from ._attr import Attr
from ._cache import Cache
from ._client_app import ClientApp
from ._counter import Counter
from ._default import Default
from ._diff import Diff
from ._encoders import app_encoder
from ._extn import Extn
from ._func import Func, Perm
from ._job import Job
from ._l10n import L10N
from ._module import Module
from ._package import App, Env, Package
from ._query import Query, QueryItem, QueryMod
from ._sys_doc import SysDoc
from ._user_attr import UserAttr
from ._var import Var

__all__ = [
    "Analytics",
    "Attr",
    "Cache",
    "ClientApp",
    "Counter",
    "Default",
    "Diff",
    "app_encoder",
    "Extn",
    "Func",
    "Perm",
    "Job",
    "L10N",
    "Module",
    "App",
    "Env",
    "Package",
    "Query",
    "QueryItem",
    "QueryMod",
    "SysDoc",
    "UserAttr",
    "Var",
]



"""Provides 'extract_val', 'set_val', 'expand_val' Utilities"""

import datetime
import logging
import os
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    MutableMapping,
    MutableSequence,
    Union,
    cast,
)

from nawah.classes._jsonpath import JSONPath
from nawah.config import Config
from nawah.enums import VarType
from nawah.exceptions import (
    InvalidLocaleException,
    InvalidVarException,
    JSONPathNotFoundException,
)

if TYPE_CHECKING:
    from nawah.classes import Var
    from nawah.types import NawahDoc, NawahEnv


logger = logging.getLogger("nawah")


def var_value(
    var: "Var",
    /,
    *,
    env: "NawahEnv" = None,
    doc: "NawahDoc" = None,
    locale: str = None,
    return_type: Union[type | Callable[[Any], Any]] = None,
) -> Any:
    """Retrieves current value of 'Var' object"""

    try:
        value = None
        match var.type:
            case VarType.ENV:
                var.var = cast(str, var.var)
                value = os.environ[var.var]

            case VarType.SESSION:
                var.var = cast(str, var.var)
                if not env:
                    raise Exception(
                        "Can't retrieve value for Var object of type 'SESSION' with "
                        "no 'env' arg provided"
                    )
                value = extract_val(scope=cast(Any, env["session"]), path=var.var)

            case VarType.DOC:
                var.var = cast(str, var.var)
                if not doc:
                    raise Exception(
                        "Can't retrieve value for Var object of type 'DOC' with "
                        "no 'doc' arg provided"
                    )
                value = extract_val(scope=cast(Any, doc), path=var.var)

            case VarType.CONFIG:
                var.var = cast(str, var.var)
                value = extract_val(scope=Config.vars, path=var.var)

            case VarType.L10N:
                var.var = cast(str, var.var)
                if not locale:
                    raise Exception(
                        "Can't retrieve value for Var object of type 'L10N' with "
                        "no 'locale' arg provided"
                    )
                if locale not in Config.locales:
                    raise InvalidLocaleException(locale=locale)
                value = extract_val(scope=Config.l10n[locale], path=var.var)

            case VarType.DATE:
                return datetime.datetime.utcnow().isoformat().split("T")[0]

            case VarType.TIME:
                return datetime.datetime.utcnow().isoformat().split("T")[1]

            case VarType.DATETIME:
                return datetime.datetime.utcnow().isoformat()

            case _:
                raise Exception(f"Unknown Var Type '{var.type}'")

        if not return_type:
            return value

        return return_type(value)

    except (KeyError, ValueError) as e:
        raise InvalidVarException(var=var) from e


def _jsonpath_eval_func(*args):
    return eval(*args)


def extract_val(
    *, scope: MutableMapping[str, Any], path: str, return_array: bool = False
):
    """Extracts value from provided 'scope' with JSONPath-based 'path'"""

    if path.startswith("$__"):
        path = path[3:]

    attr = JSONPath(path).parse(scope, eval_func=_jsonpath_eval_func)

    if return_array:
        return attr

    if len(attr) == 0:
        raise JSONPathNotFoundException(scope=scope, attr_path=path)

    return attr[0]


def set_val(*, scope: dict[str, Any], path: str, value: Any):
    """Sets value to object from provided 'scope' with JSONPath-based 'path'"""

    if path.startswith("$__"):
        path = path[3:]

    # Find the path to target parent
    attr_path_parent = ".".join(path.split(".")[:-1])

    # Set 'attr' to list containing 'scope', to have value for attr
    # if 'attr_path_parent' is empty, skipping JSONPath parse
    attr = [scope]

    if attr_path_parent:
        attr = JSONPath(attr_path_parent).parse(scope, eval_func=_jsonpath_eval_func)

    if len(attr) == 0:
        raise JSONPathNotFoundException(scope=scope, attr_path=path)

    attr_path_last: Union[str, int] = path.split(".")[-1]
    if re.match(r"^[0-9]$", cast(str, attr_path_last)):
        attr_path_last = int(attr_path_last)

    for child_attr in attr:
        child_attr[
            attr_path_last if isinstance(child_attr, list) else str(attr_path_last)
        ] = value


def expand_val(
    *, doc: MutableMapping[str, Any], expanded_doc: MutableMapping[str, Any] = None
):
    """Expands dot-notated keys in dict to its full path"""

    if not expanded_doc:
        expanded_doc = {}
    for attr in doc:
        if isinstance(doc[attr], dict):
            doc[attr] = expand_val(doc=doc[attr])
        if "." in attr:
            attr_path = attr.split(".")
            scope = expanded_doc
            for i in range(len(attr_path) - 1):
                try:
                    if not isinstance(scope[attr_path[i]], dict):
                        scope[attr_path[i]] = {}
                except KeyError:
                    scope[attr_path[i]] = {}
                scope = scope[attr_path[i]]
            scope[attr_path[-1]] = doc[attr]
        else:
            expanded_doc[attr] = doc[attr]
    return expanded_doc


def deep_update(
    *,
    target: Union[MutableMapping, MutableSequence],
    new_values: Union[MutableMapping, MutableSequence],
):
    """Updates values from two dics, lists recursively down to the last child"""

    if not isinstance(target, type(new_values)):
        # [TODO] Refactor as custom exception
        raise Exception(
            f"Type '{type(target)}' of 'target' is not the same as '{type(new_values)}'"
            "of 'new_values'",
        )

    if isinstance(new_values, dict):
        new_values = cast(dict, new_values)
        for k in new_values.keys():
            target = cast(dict, target)
            if k not in target.keys():
                target[k] = new_values[k]
            else:
                deep_update(target=target[k], new_values=new_values[k])
    elif isinstance(new_values, list):
        for j in new_values:
            target = cast(list, target)
            if j not in target:
                target.append(j)


def camel_to_upper(val: str, /) -> str:
    """Converts 'val' from CamelCase to UPPER_CASE"""

    # [REF] https://stackoverflow.com/a/1176023
    return re.sub(r"(?<!^)(?=[A-Z])", "_", val).upper()



"""Provides 'Module' dataclass"""

from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Callable,
    MutableMapping,
    MutableSequence,
    Optional,
    Tuple,
    Union,
)

if TYPE_CHECKING:
    from ._attr import Attr
    from ._counter import Counter
    from ._default import Default
    from ._diff import Diff
    from ._extn import Extn
    from ._func import Func


@dataclass(kw_only=True)
class Module:
    """Module dataclass serves role of defining a Nawah Datatype along the
    funcs required for it. It is at the centre of Nawah, as in, any call
    in Nawah is directed to a specific Module and fulfilled by one of its
    defined funcs, completed with other configurations defined."""

    # pylint: disable=too-many-instance-attributes

    name: str
    funcs: MutableMapping[str, "Func"]
    post_config: Optional[Callable[[], None]] = None
    collection: Optional[str] = None
    desc: Optional[str] = None
    attrs: MutableMapping[str, "Attr"] = field(default_factory=lambda: {})
    unique_attrs: MutableSequence[Union[Tuple[str, ...], str]] = field(
        default_factory=lambda: []
    )
    counters: MutableMapping[str, "Counter"] = field(default_factory=lambda: {})
    diff: Optional["Diff"] = None
    create_draft: bool = False
    update_draft: bool = False
    defaults: MutableMapping[str, "Default"] = field(default_factory=lambda: {})
    extns: MutableMapping[str, "Extn"] = field(default_factory=lambda: {})
    privileges: MutableSequence[str] = field(
        default_factory=lambda: [
            "admin",
            "read",
            "create",
            "update",
            "delete",
        ]
    )



"""Provides 'Func', 'Perm' dataclasses"""

from dataclasses import dataclass
from typing import (TYPE_CHECKING, Awaitable, Callable, MutableMapping,
                    MutableSequence, Optional, Type, Union)

if TYPE_CHECKING:
    from nawah.types import NawahDoc, Results

    from ._analytics import Analytics
    from ._attr import Attr
    from ._cache import Cache
    from ._module import Module
    from ._query import QueryMod


@dataclass(kw_only=True)
class Func:
    """Func dataclass serves role of defining a Nawah Module Function, which identifies set of
    configuration to expose a callable as an endpoint"""

    # pylint: disable=too-many-instance-attributes

    permissions: MutableSequence["Perm"]
    analytics: Optional["Analytics"] = None
    exceptions: Optional[MutableMapping[Type[Exception], bool]] = None
    cache: Optional["Cache"] = None
    cache_channels_update: Optional[MutableSequence[str]] = None
    cache_channels_reset: Optional[MutableSequence[str]] = None
    call_args: Optional[MutableMapping[str, "Attr"]] = None
    query_attrs: Optional[
        Union[MutableMapping[str, "Attr"], MutableSequence[MutableMapping[str, "Attr"]]]
    ] = None
    doc_attrs: Optional[
        Union[MutableMapping[str, "Attr"], MutableSequence[MutableMapping[str, "Attr"]]]
    ] = None
    strict_doc: bool = False
    get_func: bool = False
    post_func: bool = False
    callable: Optional[Callable[..., Awaitable["Results"]]] = None
    desc: Optional[str] = None

    @property
    def name(self) -> str:
        """Getter for '_name' attr"""
        return self._name

    @name.setter
    def name(self, val: str):
        """Setter for '_name' attr"""
        self._name = val

    @property
    def module(self) -> "Module":
        """Getter for '_module' attr"""
        return self._module

    @module.setter
    def module(self, val: "Module"):
        """Setter for '_module' attr"""
        self._module = val

    def __post_init__(self):
        # Create private instance attributes
        self._name = None
        self._module = None
        # Convert query_attrs, doc_attrs to list if defined as dict
        if isinstance(self.query_attrs, dict):
            self.query_attrs = [self.query_attrs]
        if isinstance(self.doc_attrs, dict):
            self.doc_attrs = [self.doc_attrs]


@dataclass(kw_only=True)
class Perm:
    """Perm dataclass serves role of defining a Permissions Set, where Nawah attempts to match
    value of 'privilege' with current 'User' privileges, and use values for 'query_mod', 'doc_mod'
    to modify respective values"""

    privilege: str
    query_mod: Optional["QueryMod"] = None
    doc_mod: Optional["NawahDoc"] = None



"""Provides 'Attr' dataclass"""

import datetime
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Union,
)

from nawah.enums import AttrType

if TYPE_CHECKING:
    from ._default import Default
    from ._extn import Extn


@dataclass(kw_only=True)
class Attr:
    """Attr dataclass serves role of defining an Attr Type Set, which
    instructs Nawah of allowed data type for a specific value"""

    # pylint: disable=invalid-name, redefined-builtin, too-many-public-methods

    desc: Optional[str]
    type: AttrType
    args: MutableMapping[str, Any]

    def __post_init__(self):
        self._default = None
        self._extn = None
        self._valid = False

    @property
    def default(self) -> Optional["Default"]:
        """Exposes _default attribute"""
        return self._default

    @default.setter
    def default(self, default):
        self._default = default

    @property
    def extn(self) -> Optional["Extn"]:
        """Exposes _extn attribute"""
        return self._extn

    @extn.setter
    def extn(self, extn):
        self._extn = extn

    @property
    def valid(self) -> bool:
        """Exposes _valid attribute"""
        return self._valid

    @valid.setter
    def valid(self, valid):
        self._valid = valid

    @staticmethod
    def ANY(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'ANY'"""
        return Attr(desc=desc, type=AttrType.ANY, args={})

    @staticmethod
    def ATTR(desc: str = None, type: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'ATTR'"""
        return Attr(desc=desc, type=AttrType.ATTR, args={"type": type})

    @staticmethod
    def ID(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'ID'"""
        return Attr(desc=desc, type=AttrType.ID, args={})

    @staticmethod
    def STR(pattern: str = None, desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'STR'"""
        return Attr(desc=desc, type=AttrType.STR, args={"pattern": pattern})

    @staticmethod
    def INT(ranges: Sequence[Sequence[float]] = None, desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'INT'"""
        return Attr(desc=desc, type=AttrType.INT, args={"ranges": ranges})

    @staticmethod
    def FLOAT(ranges: Sequence[Sequence[float]] = None, desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'FLOAT'"""
        return Attr(desc=desc, type=AttrType.FLOAT, args={"ranges": ranges})

    @staticmethod
    def BOOL(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'BOOL'"""
        return Attr(desc=desc, type=AttrType.BOOL, args={})

    @staticmethod
    def LOCALE(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'LOCALE'"""
        return Attr(desc=desc, type=AttrType.LOCALE, args={})

    @staticmethod
    def LOCALES(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'LOCALES'"""
        return Attr(desc=desc, type=AttrType.LOCALES, args={})

    @staticmethod
    def EMAIL(
        allowed_domains: Sequence[str] = None,
        disallowed_domains: Sequence[str] = None,
        strict_matching: bool = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'EMAIL'"""
        return Attr(
            desc=desc,
            type=AttrType.EMAIL,
            args={
                "allowed_domains": allowed_domains,
                "disallowed_domains": disallowed_domains,
                "strict_matching": strict_matching,
            },
        )

    @staticmethod
    def PHONE(codes: Sequence[str] = None, desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'PHONE'"""
        return Attr(desc=desc, type=AttrType.PHONE, args={"codes": codes})

    @staticmethod
    def IP(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'IP'"""
        return Attr(desc=desc, type=AttrType.IP, args={})

    @staticmethod
    def URI_WEB(
        allowed_domains: Sequence[str] = None,
        disallowed_domains: Sequence[str] = None,
        strict_matching: bool = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'URI_WEB'"""
        return Attr(
            desc=desc,
            type=AttrType.URI_WEB,
            args={
                "allowed_domains": allowed_domains,
                "disallowed_domains": disallowed_domains,
                "strict_matching": strict_matching,
            },
        )

    @staticmethod
    def DATETIME(
        ranges: Sequence[Sequence[datetime.datetime]] = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'DATETIME'"""
        return Attr(desc=desc, type=AttrType.DATETIME, args={"ranges": ranges})

    @staticmethod
    def DATE(
        ranges: Sequence[Sequence[datetime.date]] = None, desc: str = None
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'DATE'"""
        return Attr(desc=desc, type=AttrType.DATE, args={"ranges": ranges})

    @staticmethod
    def TIME(ranges: Sequence[Sequence[str]] = None, desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'TIME'"""
        return Attr(desc=desc, type=AttrType.TIME, args={"ranges": ranges})

    @staticmethod
    def FILE(
        types: Sequence[str] = None,
        ratio_ranges: Sequence[Sequence[float]] = None,
        dims_ranges: Sequence[Sequence[float]] = None,
        size_ranges: Sequence[Sequence[float]] = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'FILE'"""
        return Attr(
            desc=desc,
            type=AttrType.FILE,
            args={
                "types": types,
                "ratio_ranges": ratio_ranges,
                "dims_ranges": dims_ranges,
                "size_ranges": size_ranges,
            },
        )

    @staticmethod
    def GEO_POINT(desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'GEO_POINT'"""
        return Attr(desc=desc, type=AttrType.GEO_POINT, args={})

    @staticmethod
    def LIST(
        list: Sequence["Attr"],
        len_range: Sequence[float] = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'LIST'"""
        return Attr(
            desc=desc, type=AttrType.LIST, args={"list": list, "len_range": len_range}
        )

    @staticmethod
    def KV_DICT(
        key: "Attr",
        val: "Attr",
        req: Sequence[str] = None,
        len_range: Sequence[float] = None,
        desc: str = None,
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'KV_DICT'"""
        return Attr(
            desc=desc,
            type=AttrType.KV_DICT,
            args={"key": key, "val": val, "req": req, "len_range": len_range},
        )

    @staticmethod
    def TYPED_DICT(dict: Mapping[str, "Attr"], desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'TYPED_DICT'"""
        return Attr(desc=desc, type=AttrType.TYPED_DICT, args={"dict": dict})

    @staticmethod
    def LITERAL(
        literal: Sequence[Union[str, int, float, bool]], desc: str = None
    ) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'LITERAL'"""
        return Attr(desc=desc, type=AttrType.LITERAL, args={"literal": literal})

    @staticmethod
    def UNION(union: Sequence["Attr"], desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'UNION'"""
        return Attr(desc=desc, type=AttrType.UNION, args={"union": union})

    @staticmethod
    def TYPE(type: Union[Callable, str], desc: str = None) -> "Attr":
        """Shorthand method to define 'Attr' object of type 'TYPE'"""
        return Attr(desc=desc, type=AttrType.TYPE, args={"type": type})



"""Provides 'Config', 'ConfigAttrSys' static classes"""

from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Set, Type, TypedDict)

from nawah.enums import LocaleStrategy

if TYPE_CHECKING:
    from datetime import datetime

    import redis.commands.json.JSON
    from bson import ObjectId
    from motor.motor_asyncio import AsyncIOMotorClient

    from nawah.classes import (L10N, Attr, ClientApp, Job, Module, SysDoc,
                               UserAttr)
    from nawah.types import (AnalyticsEvents, AppPackage, IPQuota, NawahDoc,
                             NawahEnv)


class VarTypeDict(TypedDict):
    """Provides type-hint for item of 'vars_types' runtime config"""

    package: str
    type: "Attr"


class ConfigAttrSys:
    """ConfigAttrSys serves role of static class used as value for 'sys' attribute
    of 'Config' static class"""

    # pylint: disable=too-few-public-methods

    admin_check: bool = False

    conn: "AsyncIOMotorClient"
    cache: Optional["redis.commands.json.JSON"] = None
    env: "NawahEnv"
    docs: MutableMapping["ObjectId", "SysDoc"] = {}
    jobs_base: "datetime"

    nawah_version: str

    name: str
    version: str
    path: str
    packages: MutableMapping[str, "AppPackage"] = {}

    cache_channels: Set[str] = set()

    sessions: MutableMapping[str, "NawahEnv"] = {}
    session_counter = 0
    session_lock: Optional[float] = None
    ip_quota: MutableMapping[str, "IPQuota"] = {}

    type_attrs: MutableSequence["Attr"] = []


class Config:
    """Config serves role of static class used as central reference for runtime
    config of Nawah App"""

    # pylint: disable=too-few-public-methods

    debug: bool = False
    env: str
    port: int = 8081

    sys: Type[ConfigAttrSys] = ConfigAttrSys

    test: bool = False

    emulate_test: bool = False

    vars_types: MutableMapping[str, "VarTypeDict"] = {}
    vars: MutableMapping[str, Any] = {}

    client_apps: MutableMapping[str, "ClientApp"] = {}

    analytics_events: "AnalyticsEvents" = {
        "app_conn_verified": True,
        "session_conn_auth": True,
        "session_user_auth": True,
        "session_conn_reauth": True,
        "session_user_reauth": True,
        "session_conn_deauth": True,
        "session_user_deauth": True,
    }

    conn_timeout: int = 120
    quota_anon_min: int = 40
    quota_auth_min: int = 100
    quota_ip_min: int = 500
    file_upload_limit: int = -1
    file_upload_timeout: int = 300

    data_server: str = "mongodb://localhost"
    data_name: str = "nawah_data"
    data_ssl: bool = False
    data_disk_use: bool = False

    cache_server: Optional[str] = None
    cache_db: Optional[int] = 0
    cache_username: Optional[str] = None
    cache_password: Optional[str] = None
    cache_expiry: Optional[int] = None

    error_reporting_server: Optional[str] = None

    locales: MutableSequence[str] = ["ar_AE", "en_GB"]
    locale: str = "ar_AE"
    locale_strategy: "LocaleStrategy" = LocaleStrategy.DUPLICATE

    admin_doc: "NawahDoc" = {}
    admin_password: str

    anon_token: str
    anon_privileges: MutableMapping[str, MutableSequence[str]] = {}
    anon_doc: Optional["NawahDoc"] = None

    auth_attrs: MutableSequence[str] = []
    user_attrs_sanitise: MutableSequence[str] = []
    user_attrs: MutableMapping[str, "UserAttr"] = {}

    groups: MutableSequence[MutableMapping[str, Any]] = []
    default_privileges: MutableMapping[str, MutableSequence[str]] = {}

    data_indexes: MutableSequence[MutableMapping[str, Any]] = []

    docs: MutableSequence["SysDoc"] = []

    l10n: MutableMapping[str, "L10N"] = {}

    jobs: MutableMapping[str, "Job"] = {}

    types: MutableMapping[str, Callable] = {}

    modules: MutableMapping[str, "Module"] = {}



"""Provides 'settings' Nawah Core Module"""

from nawah.classes import Attr, Diff, Extn, Func, Module, Perm, Var

settings = Module(
    name="settings",
    desc="'settings' module module provides data type and controller for settings in Nawah "
    "eco-system. This reflects stateful config that are needed for Nawah App to operate",
    collection="setting_docs",
    attrs={
        "var": Attr.STR(
            desc="Name of the setting. This is unique for every 'user' in the module"
        ),
        "val": Attr.ANY(desc="Value of the setting"),
        "val_type": Attr.ATTR(),
    },
    diff=Diff(condition=lambda *_: True),
    unique_attrs=["var"],
    extns={
        "val": Extn(
            module=Var.DOC("val.__extn.module"),
            attrs=Var.DOC("val.__extn.attrs"),
            force=Var.DOC("val.__extn.force"),
        ),
    },
    funcs={
        "read": Func(
            permissions=[Perm(privilege="read")],
        ),
        "create": Func(
            permissions=[
                Perm(privilege="create"),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "update": Func(
            permissions=[
                Perm(privilege="update", query_mod={"$limit": 1}),
            ],
            query_attrs=[{"_id": Attr.ID()}, {"var": Attr.STR()}],
            call_args={"raise_no_success": Attr.BOOL()},
        ),
        "delete": Func(
            permissions=[Perm(privilege="delete", query_mod={"$limit": 1})],
            query_attrs=[{"_id": Attr.ID()}, {"var": Attr.STR()}],
        ),
        "retrieve_file": Func(
            permissions=[Perm(privilege="*")],
            get_func=True,
        ),
    },
)



"""Provides 'retrieve_file' Base Function callable"""

import io
import logging
from typing import (TYPE_CHECKING, Any, MutableMapping, MutableSequence,
                    Optional, cast)

from bson import ObjectId
from PIL import Image

from nawah import data as Data
from nawah.classes import Query
from nawah.config import Config

from ._shared import FILE_ATTRS
from .exceptions import (FileNotFoundException, FileNotImageException,
                         InvalidDocException, UtilityModuleDataCallException)

if TYPE_CHECKING:
    from nawah.types import NawahEnv, Results

logger = logging.getLogger("nawah")


async def retrieve_file(
    *,
    module_name: str,
    env: "NawahEnv",
    query: "Query",
) -> "Results":
    """Finds file from doc in module, and returns it as part of results. If doc not found, raises
    'InvalidDocException'. If attr not found in doc, of file not matching file[s] in attr, raises
    'FileNotFoundException'. If a value for arg 'thumb_dims' is provided, generates a thumbnail, but
    raises 'FileNotImageException' if file not an image. If process to generate thumbnail fails, for
    any reasons, original file would be returned, error would be logged"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="retrieve_file"
        )

    attr_name = query["attr:$eq"][0]
    filename = query["filename:$eq"][0]
    thumb_dims: Optional[MutableSequence[int]] = None
    if "thumb:$eq" in query:
        thumb_dims = [int(dim) for dim in query["thumb:$eq"][0].split("x")]

    file_results = await Data.read(
        env=env,
        collection_name=f"{module.collection}__file",
        attrs=FILE_ATTRS,
        query=Query([{"_id": {"$eq": ObjectId(query["_id:$eq"][0])}}]),
    )

    if not file_results["count"]:
        raise InvalidDocException(doc_id=query["_id:$eq"][0])

    retrieved_file = file_results["docs"][0]["file"]

    retrieved_file["content"] = _generate_thumb(
        thumb_dims=thumb_dims,
        module_name=module_name,
        doc_id=query["_id:$eq"][0],
        attr_name=attr_name,
        filename=filename,
        retrieved_file=retrieved_file,
    )

    return {
        "status": 200,
        "msg": "File attached to response",
        "args": {
            "return": "file",
            "docs": [
                {
                    "_id": query["_id:$eq"][0],
                    "name": retrieved_file["name"],
                    "type": retrieved_file["type"],
                    "lastModified": retrieved_file["lastModified"],
                    "size": retrieved_file["size"],
                    "content": retrieved_file["content"],
                }
            ],
        },
    }


def _retrieve_file_from_attr(*, doc_id, attr_name, attr, filename):
    """Attempts to find file in attr. if failed, raises 'FileNotFoundException'"""

    retrieved_file = None

    if isinstance(attr, list):
        for item in attr:
            if item["name"] == filename:
                retrieved_file = item
                break
    elif isinstance(attr, dict):
        attr = cast(MutableMapping[str, Any], attr)
        if attr["name"] == filename:
            retrieved_file = attr

    if not retrieved_file:
        # [DOC] No filename match
        raise FileNotFoundException(
            doc_id=doc_id, attr_name=attr_name, file_name=filename
        )

    return retrieved_file


def _generate_thumb(
    *, thumb_dims, module_name, doc_id, attr_name, filename, retrieved_file
):
    """Checks 'thumb_dims' value to generate thumbnail for image. if file is not image, raises
    'FileNotImageException'. If failed while generating thumbnail, logs error, and returns
    original file"""

    if not thumb_dims:
        return retrieved_file["content"]

    if retrieved_file["type"].split("/")[0] != "image":
        raise FileNotImageException(
            doc_id=doc_id,
            attr_name=attr_name,
            file_name=filename,
            file_type=retrieved_file["type"],
        )
    try:
        image = Image.open(io.BytesIO(retrieved_file["content"]))
        image.thumbnail(thumb_dims)
        stream = io.BytesIO()
        image.save(stream, format=image.format)
        stream.seek(0)
        return stream.read()
    except Exception as e:  # pylint: disable=broad-except
        logger.error("Failed to generate thumbnail for:")
        logger.error("- Module     : '%s'", module_name)
        logger.error("- Doc '_id': '%s'", doc_id)
        logger.error("- Attr       : '%s'", attr_name)
        logger.error("- File       : '%s'", filename)
        logger.error("- Exception message: %s", e)

    return retrieved_file["content"]



"""Provides Base Functions for Nawah modules"""

from ._create import create
from ._create_file import create_file
from ._delete import delete
from ._delete_file import delete_file
from ._delete_lock import delete_lock
from ._obtain_lock import obtain_lock
from ._read import read
from ._read_diff import read_diff
from ._retrieve_file import retrieve_file
from ._update import update
from ._update_file import update_file

__all__ = [
    "create",
    "create_file",
    "delete",
    "delete_lock",
    "delete_file",
    "obtain_lock",
    "read",
    "read_diff",
    "retrieve_file",
    "update",
    "update_file",
]



"""Provides 'obtain_lock' Base Function callable"""

import asyncio
import logging
from typing import TYPE_CHECKING

from pymongo.errors import DuplicateKeyError

import nawah.data as Data
from nawah.classes import Query
from nawah.config import Config
from nawah.enums import DeleteStrategy

from .exceptions import FailedDeleteLockException, FailedObtainLockException

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, Results

logger = logging.getLogger("nawah")


async def obtain_lock(
    *, module_name: str, env: "NawahEnv", doc: "NawahDoc"
) -> "Results":
    """Creates a lock for a module by creating a doc in lock collection and confirms
    it is the top lock doc. If not, deletes doc and raises 'FailedObtainLockException'.
    If deleting doc failed, raises 'FailedDeleteLockException'"""

    module = Config.modules[module_name]

    if "total_attempts" not in doc:
        doc["total_attempts"] = doc["attempts"]

    logger.debug("Attempting to obtain lock for '%s'", module_name)
    try:
        create_results = await Data.create(
            env=env,
            collection_name=f"{module.collection}__lock",
            doc=doc,
        )
        logger.debug(
            "Obtained intermediate lock '%s' for '%s'",
            create_results["docs"][0]["_id"],
            module_name,
        )
        read_results = await Data.read(
            env=env,
            collection_name=f"{module.collection}__lock",
            attrs={},
            query=Query(
                [
                    {"tags": {"$in": doc["tags"]}},
                ]
                if doc["tags"]
                else [],
                special={"$sort": {"_id": 1}},
            ),
        )
        logger.debug("Attempting to read current lock for '%s'", module_name)

        if create_results["docs"][0]["_id"] == read_results["docs"][0]["_id"]:
            logger.debug("Successfully obtained lock for '%s'", module_name)
            return {
                "status": 200,
                "msg": f"Obtained lock for '{module_name}'",
                "args": create_results,
            }
    except DuplicateKeyError:
        create_results = read_results = {"count": 0, "docs": [{"_id": None}]}

    logger.warning(
        "Failed to obtain lock for '%s', with with created lock_id: '%s', but read '%s'",
        module_name,
        create_results["docs"][0]["_id"],
        read_results["docs"][0]["_id"],
    )

    if create_results["docs"][0]["_id"]:
        logger.warning(
            "Deleting lock '%s', for '%s'",
            create_results["docs"][0]["_id"],
            module_name,
        )
        lock_delete_results = await Data.delete(
            env=env,
            collection_name=f"{module.collection}__lock",
            docs=[create_results["docs"][0]["_id"]],
            strategy=DeleteStrategy.FORCE_SYS,
        )

        if lock_delete_results["count"] != 1:
            logger.error(
                "Failed to delete failed lock '%s' for '%s'. DELETE MANUALLY NOW",
                create_results["docs"][0]["_id"],
                module_name,
            )
            raise FailedDeleteLockException(
                module_name=module_name, lock_id=create_results["docs"][0]["_id"]
            )

    if doc["attempts"]:
        await asyncio.sleep(0.2)
        logger.warning("Reattempting to obtain lock for '%s'", module_name)
        doc["attempts"] -= 1
        return await obtain_lock(module_name=module_name, env=env, doc=doc)

    raise FailedObtainLockException(module_name=module_name)



"""Provides 'read' Base Function callable"""

import logging
from typing import TYPE_CHECKING, MutableSequence, Optional

import nawah.data as Data
from nawah.config import Config

from .exceptions import NoDocFoundException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, Results

logger = logging.getLogger("nawah")


async def read(
    *,
    module_name: str,
    env: "NawahEnv",
    query: "Query",
    raise_no_success: Optional[bool],
) -> "Results":
    """Reads docs from module collection that match query"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="create"
        )

    extn_list: MutableSequence[str] = []

    extn_list = [attr_name.split(".")[0] for attr_name in module.extns]

    if "$extn" in query:
        if query["$extn"] is False:
            extn_list = []
        elif isinstance(query["$extn"], list):
            extn_list = [
                attr_name
                for attr_name in query["$extn"]
                if attr_name.split(":")[0].split(".")[0] in extn_list
            ]

    if "$attrs" in query:
        extn_list = [
            attr_name for attr_name in extn_list if attr_name in query["$attrs"]
        ]

    results = await Data.read(
        env=env,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        extn_attrs=extn_list,
    )

    # [DOC] if $attrs query arg is present return only required keys.
    if "$attrs" in query:
        query["$attrs"].insert(0, "_id")
        for i in range(len(results["docs"])):
            results["docs"][i] = {
                attr: results["docs"][i][attr]
                for attr in query["$attrs"]
                if attr in results["docs"][i]
            }

    if raise_no_success is True and results["count"] == 0:
        raise NoDocFoundException(module_name=module_name)

    return {"status": 200, "msg": f'Found {results["count"]} docs', "args": results}



"""Provides 'Extn' dataclass"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, MutableSequence, Optional, Union

if TYPE_CHECKING:
    from nawah.classes import Var
    from nawah.enums import Event
    from nawah.types import NawahQuery


@dataclass(kw_only=True)
class Extn:
    """Extn dataclass serves role of defining Extension Instruction, which
    instructs Nawah to extend specific attr value unto another doc"""

    module: Union[str, "Var"]
    attrs: Union[MutableSequence[str], "Var"]
    force: Union[bool, str, "Var"] = False
    skip_events: Optional[Union[MutableSequence["Event"], "Var"]] = None
    query: Optional[Union["NawahQuery", "Var"]] = None

    def __post_init__(self):
        pass



"""Provides exceptions related to Attr Type"""


class InvalidAttrTypeException(Exception):
    """Raied by 'validate_type' Utility if Attr Type object is not valid"""

    def __init__(self, *, attr_type):
        super().__init__(f"Unknown or invalid Attr Type '{attr_type}'")


class InvalidAttrTypeRefException(Exception):
    """Raied by 'validate_type' Utility if Attr Type 'TYPE' object referes to invalid type"""

    def __init__(self, *, attr_type):
        super().__init__(f"Attr Type '{attr_type}' refers to invalid 'type'")


class InvalidAttrTypeArgException(Exception):
    """Raised by 'validate_arg' Utility if Attr Type Arg value is invalid"""

    def __init__(self, *, arg_name: str, arg_type, arg_val):
        super().__init__(
            f"Invalid Attr Type Arg for '{arg_name}' expecting type '{arg_type}' but got "
            f"'{arg_val}'."
        )


class JSONPathNotFoundException(Exception):
    """Raised by \'extract_attr\' Utility if provided didn't match any attr in 'scope'"""

    status = 400

    def __init__(self, *, scope, attr_path):
        super().__init__(f"Failed to extract {attr_path} from scope: {scope}")



"""Provides enums required for Nawah Fremework and apps"""

from enum import Enum, auto


class AttrType(Enum):
    """Enum for all Nawah-defined Attr Types"""

    ANY = auto()
    ATTR = auto()
    ID = auto()
    STR = auto()
    INT = auto()
    FLOAT = auto()
    BOOL = auto()
    LOCALE = auto()
    LOCALES = auto()
    EMAIL = auto()
    PHONE = auto()
    IP = auto()
    URI_WEB = auto()
    DATETIME = auto()
    DATE = auto()
    TIME = auto()
    FILE = auto()
    BYTES = auto()
    GEO_POINT = auto()
    LIST = auto()
    KV_DICT = auto()
    TYPED_DICT = auto()
    LITERAL = auto()
    UNION = auto()
    TYPE = auto()


class Event(Enum):
    """Identifies events in calls that can be skipped"""

    ATTRS_QUERY = auto()
    ATTRS_DOC = auto()
    VALIDATE = auto()
    PERM = auto()
    CACHE = auto()
    SOFT = auto()
    DIFF = auto()
    SYS_DOCS = auto()
    ANALYTICS = auto()


class DeleteStrategy(Enum):
    """Identifies strategies of delete calls"""

    SOFT_SKIP_SYS = auto()
    SOFT_SYS = auto()
    FORCE_SKIP_SYS = auto()
    FORCE_SYS = auto()


class LocaleStrategy(Enum):
    """Identifies strategies of values for attrs of Attr Type LOCALE"""

    DUPLICATE = auto()
    NONE_VALUE = auto()


class NawahValues(Enum):
    """Identifies values to be used internally to pass secure values"""

    NONE_VALUE = auto()
    ALLOW_MOD = auto()


class VarType(Enum):
    """Enum for variables types to be used with 'Var' dataclass"""

    ENV = auto()
    SESSION = auto()
    DOC = auto()
    L10N = auto()
    CONFIG = auto()
    DATE = auto()
    TIME = auto()
    DATETIME = auto()



"""Provides 'session' Module Functions callables"""

import datetime
import secrets
from typing import TYPE_CHECKING, Optional

from passlib.hash import pbkdf2_sha512

from nawah.classes import Query
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call

from ._exceptions import (AnonReauthException, AnonSignoutException,
                          ExpiredSessionException, InvalidCredentialsException,
                          InvalidSessionException, InvalidUserException)

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, Results


async def _auth(
    env: "NawahEnv", doc: "NawahDoc", skip_status_check: Optional[None]
) -> "Results":
    key: str
    for attr in Config.modules["user"].unique_attrs:
        if isinstance(attr, str) and attr in doc:
            key = attr
            break

    user_query = Query([{key: {"$eq": doc[key]}}], special={"$limit": 1})

    if "groups" in doc and doc["groups"]:
        user_query.append({"groups": {"$in": doc["groups"]}})
        user_query.append({"privileges": {"$eq": {"*": ["*"]}}})

    user_results = await call(
        "user/read",
        skip_events=[Event.PERM],
        env=env,
        query=user_query,
        args={"skip_sanitise_results": True},
    )

    if not user_results["args"]["count"] or not pbkdf2_sha512.verify(
        doc["hash"],
        user_results["args"]["docs"][0][f"{key}_hash"],
    ):
        raise InvalidCredentialsException()

    user = user_results["args"]["docs"][0]

    if skip_status_check is not True:
        if user["status"] in ["banned", "deleted", "disabled_password"]:
            raise InvalidUserException(reason=user["status"])

    token = secrets.token_urlsafe(32)
    session = {
        "user": user["_id"],
        "groups": doc["groups"] if "groups" in doc.keys() else [],
        "host_add": env["REMOTE_ADDR"],
        "user_agent": env["HTTP_USER_AGENT"],
        "expiry": (
            datetime.datetime.utcnow() + datetime.timedelta(days=30)
        ).isoformat(),
        "token_hash": pbkdf2_sha512.using(rounds=100000).hash(token),
    }

    results = await call(
        "session/create", skip_events=[Event.PERM], env=env, doc=session
    )
    if results["status"] != 200:
        return results

    session["_id"] = results["args"]["docs"][0]["_id"]
    session["user"] = user
    del session["token_hash"]
    session["token"] = token
    results["args"]["docs"][0] = session

    # read user privileges and return them
    user_results = await call(
        "user/read_privileges",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": user["_id"]}}]),
    )
    if user_results["status"] != 200:
        return user_results
    results["args"]["docs"][0]["user"] = user_results["args"]["docs"][0]

    return {
        "status": 200,
        "msg": "You were successfully authed.",
        "args": {"session": results["args"]["docs"][0]},
    }


async def _reauth(env: "NawahEnv", query: "Query") -> "Results":
    if str(query["_id:$eq"][0]) == "f00000000000000000000012":
        raise AnonReauthException()

    session_query = Query([{"_id": {"$eq": query["_id:$eq"][0]}}])
    if query["groups:$eq"][0]:
        session_query.append({"groups": {"$in": query["groups:$eq"][0]}})
    results = await call(
        "session/read", skip_events=[Event.PERM], env=env, query=session_query
    )
    if not results["args"]["count"]:
        raise InvalidSessionException()

    if not pbkdf2_sha512.verify(
        query["token:$eq"][0], results["args"]["docs"][0]["token_hash"]
    ):
        raise InvalidSessionException()

    del results["args"]["docs"][0]["token_hash"]
    results["args"]["docs"][0]["token"] = query["token:$eq"][0]

    if results["args"]["docs"][0]["expiry"] < datetime.datetime.utcnow().isoformat():
        results = await call(
            "session/delete",
            skip_events=[Event.PERM, Event.SOFT],
            env=env,
            query=Query([{"_id": {"$eq": env["session"]["_id"]}}]),
        )
        raise ExpiredSessionException()

    # update user's last_login timestamp
    await call(
        "user/update",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": results["args"]["docs"][0]["user"]["_id"]}}]),
        doc={"login_time": datetime.datetime.utcnow().isoformat()},
    )
    await call(
        "session/update",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": results["args"]["docs"][0]["_id"]}}]),
        doc={
            "expiry": (
                datetime.datetime.utcnow() + datetime.timedelta(days=30)
            ).isoformat()
        },
    )
    # read user privileges and return them
    user_results = await call(
        "user/read_privileges",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": results["args"]["docs"][0]["user"]["_id"]}}]),
    )
    results["args"]["docs"][0]["user"] = user_results["args"]["docs"][0]

    return {
        "status": 200,
        "msg": "You were successfully reauthed.",
        "args": {"session": results["args"]["docs"][0]},
    }


async def _signout(env: "NawahEnv", query: "Query") -> "Results":
    if str(query["_id:$eq"][0]) == "f00000000000000000000012":
        raise AnonSignoutException()

    results = await call(
        "session/read",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
    )

    if not results["args"]["count"]:
        raise InvalidSessionException()

    results = await call(
        "session/delete",
        skip_events=[Event.PERM],
        env=env,
        query=Query([{"_id": {"$eq": env["session"]["_id"]}}]),
    )

    return {
        "status": 200,
        "msg": "You are successfully signed-out.",
        "args": {"session": {"_id": "f00000000000000000000012"}},
    }



"""Provides exceptions related to 'var_value' Utility"""


class InvalidLocaleException(Exception):
    """Raied by 'var_value' Utility if 'locale' is not defined in runtime config"""

    def __init__(self, *, locale):
        super().__init__(f"Invalid locale '{locale}'")


class InvalidVarException(Exception):
    """Raied by 'var_value' Utility if failed to extract value of 'var'"""

    def __init__(self, *, var):
        super().__init__(f"Invalid var '{var}'")
